"""A modification workflow for a GÉANT IP subscription."""

from typing import Annotated, Any
from uuid import uuid4

from orchestrator import begin, conditional, done, step, workflow
from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, UUIDstr
from orchestrator.workflow import StepList
from orchestrator.workflows.steps import resync, store_process_subscription, unsync
from orchestrator.workflows.utils import wrap_modify_initial_input_form
from pydantic import AfterValidator, BaseModel, ConfigDict, Field, computed_field
from pydantic_forms.types import State
from pydantic_forms.validators import Divider, Label

from gso.products.product_blocks.bgp_session import BGPSession, IPFamily
from gso.products.product_blocks.geant_ip import NRENAccessPort
from gso.products.product_blocks.service_binding_port import VLAN_ID, ServiceBindingPort
from gso.products.product_types.edge_port import EdgePort
from gso.products.product_types.geant_ip import GeantIP
from gso.utils.helpers import active_edge_port_selector
from gso.utils.shared_enums import APType, SBPType
from gso.utils.types.ip_address import IPv4AddressType, IPv6AddressType


def initial_input_form_generator(subscription_id: UUIDstr) -> FormGenerator:
    """Get input about added, removed, and modified Access Ports."""
    subscription = GeantIP.from_subscription(subscription_id)

    class AccessPortSelection(BaseModel):
        geant_ip_ep: active_edge_port_selector(partner_id=subscription.customer_id)  # type: ignore[valid-type]
        nren_ap_type: APType

    def validate_edge_ports_are_unique(access_ports: list[AccessPortSelection]) -> list[AccessPortSelection]:
        """Verify if interfaces are unique."""
        edge_ports = [port.geant_ip_ep.name for port in access_ports]
        if len(edge_ports) != len(set(edge_ports)):
            msg = "Edge Ports must be unique."
            raise ValueError(msg)
        return access_ports

    class ModifyGeantIPAccessPortsForm(FormPage):
        model_config = ConfigDict(title="Modify GÉANT IP")
        access_ports: Annotated[list[AccessPortSelection], AfterValidator(validate_edge_ports_are_unique)] = [  # noqa: RUF012
            AccessPortSelection(
                geant_ip_ep=str(access_port.geant_ip_sbp.edge_port.owner_subscription_id),
                nren_ap_type=access_port.nren_ap_type,
            )
            for access_port in subscription.geant_ip.geant_ip_ap_list
        ]

    access_port_input = yield ModifyGeantIPAccessPortsForm
    input_ap_list = access_port_input.access_ports
    input_ep_list = [str(ap.geant_ip_ep) for ap in input_ap_list]
    existing_ep_list = [
        str(ap.geant_ip_sbp.edge_port.owner_subscription_id) for ap in subscription.geant_ip.geant_ip_ap_list
    ]

    class BaseBGPPeer(BaseModel):
        bfd_enabled: bool = False
        bfd_interval: int | None = None
        bfd_multiplier: int | None = None
        has_custom_policies: bool = False
        authentication_key: str
        multipath_enabled: bool = False
        send_default_route: bool = False
        is_passive: bool = False

    class IPv4BGPPeer(BaseBGPPeer):
        peer_address: IPv4AddressType
        add_v4_multicast: bool = Field(default=False, exclude=True)

        @computed_field  # type: ignore[misc]
        @property
        def families(self) -> list[IPFamily]:
            return [IPFamily.V4UNICAST, IPFamily.V4MULTICAST] if self.add_v4_multicast else [IPFamily.V4UNICAST]

    class IPv6BGPPeer(BaseBGPPeer):
        peer_address: IPv6AddressType
        add_v6_multicast: bool = Field(default=False, exclude=True)

        @computed_field  # type: ignore[misc]
        @property
        def families(self) -> list[IPFamily]:
            return [IPFamily.V6UNICAST, IPFamily.V6MULTICAST] if self.add_v6_multicast else [IPFamily.V6UNICAST]

    #  There are three possible scenarios for Edge Ports. They can be added, removed, or their relevant SBP can be
    #  modified.
    removed_ap_list = [
        access_port.subscription_instance_id
        for access_port in subscription.geant_ip.geant_ip_ap_list
        if str(access_port.geant_ip_sbp.edge_port.owner_subscription_id) not in input_ep_list
    ]
    modified_ap_list = [
        (
            access_port,
            next(
                (
                    ap.nren_ap_type
                    for ap in input_ap_list
                    if str(ap.geant_ip_ep) == str(access_port.geant_ip_sbp.edge_port.owner_subscription_id)
                ),
                None,
            ),
        )
        for access_port in subscription.geant_ip.geant_ip_ap_list
        if str(access_port.geant_ip_sbp.edge_port.owner_subscription_id) in input_ep_list
    ]
    added_ap_list = [
        (ep, next(ap.nren_ap_type for ap in input_ap_list if str(ap.geant_ip_ep) == ep))
        for ep in input_ep_list
        if ep not in existing_ep_list
    ]

    #  First, the user can modify existing Edge Ports
    sbp_inputs = []
    for access_port_index, ap_entry in enumerate(modified_ap_list):
        access_port, new_ap_type = ap_entry
        current_sbp = access_port.geant_ip_sbp
        v4_peer = next(peer for peer in current_sbp.sbp_bgp_session_list if IPFamily.V4UNICAST in peer.families)
        v6_peer = next(peer for peer in current_sbp.sbp_bgp_session_list if IPFamily.V6UNICAST in peer.families)

        class BindingPortModificationForm(FormPage):
            model_config = ConfigDict(
                title=f"GÉANT IP - Modify Edge Port configuration ({access_port_index + 1}/{len(input_ap_list)})"
            )
            current_ep_label: Label = Field(
                f"Currently configuring on {access_port.geant_ip_sbp.edge_port.description} "
                f"(Access Port type: {access_port.nren_ap_type})",
                exclude=True,
            )

            geant_sid: str = current_sbp.geant_sid
            is_tagged: bool = current_sbp.is_tagged
            # The SBP model doesn't require these three fields, but in the case of GÉANT IP this will never occur since
            # it's a layer 3 service. The ignore statements are there to put our type checker at ease.
            vlan_id: VLAN_ID = current_sbp.vlan_id  # type: ignore[assignment]
            ipv4_address: IPv4AddressType = current_sbp.ipv4_address  # type: ignore[assignment]
            ipv6_address: IPv6AddressType = current_sbp.ipv6_address  # type: ignore[assignment]
            custom_firewall_filters: bool = current_sbp.custom_firewall_filters
            divider: Divider = Field(None, exclude=True)
            v4_bgp_peer: IPv4BGPPeer = IPv4BGPPeer(
                **v4_peer.model_dump(exclude=set("families")),
                add_v4_multicast=bool(IPFamily.V4MULTICAST in v4_peer.families),
            )
            v6_bgp_peer: IPv6BGPPeer = IPv6BGPPeer(
                **v6_peer.model_dump(exclude=set("families")),
                add_v6_multicast=bool(IPFamily.V6MULTICAST in v6_peer.families),
            )

        binding_port_input_form = yield BindingPortModificationForm
        sbp_inputs.append(
            binding_port_input_form.model_dump()
            | {
                "new_ap_type": new_ap_type,
                "current_sbp_id": current_sbp.subscription_instance_id,
            }
        )

    #  Second, newly added Edge Ports are configured
    binding_port_inputs = []
    for ap_index, access_port_tuple in enumerate(added_ap_list):
        edge_port_id, ap_type = access_port_tuple

        class BindingPortInputForm(FormPage):
            model_config = ConfigDict(
                title=f"GÉANT IP - Configure new Edge Port "
                f"({len(modified_ap_list) + ap_index + 1}/{len(input_ap_list)})"
            )
            info_label: Label = Field(
                "Please configure the Service Binding Ports for each newly added Edge Port", exclude=True
            )
            current_ep_label: Label = Field(
                f"Currently configuring on {EdgePort.from_subscription(edge_port_id).description} "
                f"(Access Port type: {ap_type})",
                exclude=True,
            )

            geant_sid: str
            is_tagged: bool = False
            vlan_id: VLAN_ID
            ipv4_address: IPv4AddressType
            ipv6_address: IPv6AddressType
            custom_firewall_filters: bool = False
            divider: Divider = Field(None, exclude=True)
            v4_bgp_peer: IPv4BGPPeer
            v6_bgp_peer: IPv6BGPPeer

        binding_port_input_form = yield BindingPortInputForm
        binding_port_inputs.append(
            binding_port_input_form.model_dump()
            | {
                "bgp_peers": [
                    binding_port_input_form.v4_bgp_peer.model_dump(),
                    binding_port_input_form.v6_bgp_peer.model_dump(),
                ],
                "edge_port_id": edge_port_id,
                "ap_type": ap_type,
            }
        )

    return access_port_input.model_dump() | {
        "added_service_binding_ports": binding_port_inputs,
        "removed_access_ports": removed_ap_list,
        "modified_sbp_list": sbp_inputs,
    }


@step("Clean up removed Edge Ports")
def remove_old_sbp_blocks(subscription: GeantIP, removed_access_ports: list[UUIDstr]) -> State:
    """Remove old :term:`SBP` product blocks from the GÉANT IP subscription."""
    subscription.geant_ip.geant_ip_ap_list = [
        ap
        for ap in subscription.geant_ip.geant_ip_ap_list
        if str(ap.subscription_instance_id) not in removed_access_ports
    ]

    return {"subscription": subscription}


@step("Modify existing Service Binding Ports")
def modify_existing_sbp_blocks(subscription: GeantIP, modified_sbp_list: list[dict[str, Any]]) -> State:
    """Update the subscription model."""
    for access_port in subscription.geant_ip.geant_ip_ap_list:
        current_sbp = access_port.geant_ip_sbp
        modified_sbp_data = next(
            sbp for sbp in modified_sbp_list if sbp["current_sbp_id"] == str(current_sbp.subscription_instance_id)
        )

        v4_peer = next(peer for peer in current_sbp.sbp_bgp_session_list if IPFamily.V4UNICAST in peer.families)
        for attribute in modified_sbp_data["v4_bgp_peer"]:
            setattr(v4_peer, attribute, modified_sbp_data["v4_bgp_peer"][attribute])

        v6_peer = next(peer for peer in current_sbp.sbp_bgp_session_list if IPFamily.V6UNICAST in peer.families)
        for attribute in modified_sbp_data["v6_bgp_peer"]:
            setattr(v6_peer, attribute, modified_sbp_data["v6_bgp_peer"][attribute])

        current_sbp.sbp_bgp_session_list = [v4_peer, v6_peer]
        current_sbp.vlan_id = modified_sbp_data["vlan_id"]
        current_sbp.geant_sid = modified_sbp_data["geant_sid"]
        current_sbp.is_tagged = modified_sbp_data["is_tagged"]
        current_sbp.ipv4_address = modified_sbp_data["ipv4_address"]
        current_sbp.ipv6_address = modified_sbp_data["ipv6_address"]
        current_sbp.custom_firewall_filters = modified_sbp_data["custom_firewall_filters"]
        access_port.nren_ap_type = modified_sbp_data["new_ap_type"]

    return {"subscription": subscription}


@step("Instantiate new Service Binding Ports")
def create_new_sbp_blocks(subscription: GeantIP, added_service_binding_ports: list[dict[str, Any]]) -> State:
    """Add new :term:`SBP`s to the GÉANT IP subscription."""
    for sbp_input in added_service_binding_ports:
        edge_port = EdgePort.from_subscription(sbp_input["edge_port_id"])
        sbp_bgp_session_list = [
            BGPSession.new(subscription_id=uuid4(), **session, rtbh_enabled=True, is_multi_hop=True)
            for session in sbp_input["bgp_peers"]
        ]
        service_binding_port = ServiceBindingPort.new(
            subscription_id=uuid4(),
            **sbp_input,
            sbp_bgp_session_list=sbp_bgp_session_list,
            sbp_type=SBPType.L3,
            edge_port=edge_port.edge_port,
        )
        subscription.geant_ip.geant_ip_ap_list.append(
            NRENAccessPort.new(
                subscription_id=uuid4(),
                nren_ap_type=sbp_input["ap_type"],
                geant_ip_sbp=service_binding_port,
            )
        )

    return {"subscription": subscription}


@workflow(
    "Modify GÉANT IP",
    initial_input_form=wrap_modify_initial_input_form(initial_input_form_generator),
    target=Target.MODIFY,
)
def modify_geant_ip() -> StepList:
    """Modify a GÉANT IP subscription."""
    access_ports_are_removed = conditional(lambda state: bool(len(state["removed_access_ports"]) > 0))
    access_ports_are_modified = conditional(lambda state: bool(len(state["modified_sbp_list"]) > 0))
    access_ports_are_added = conditional(lambda state: bool(len(state["added_service_binding_ports"]) > 0))

    return (
        begin
        >> store_process_subscription(Target.MODIFY)
        >> unsync
        >> access_ports_are_removed(remove_old_sbp_blocks)
        >> access_ports_are_modified(modify_existing_sbp_blocks)
        >> access_ports_are_added(create_new_sbp_blocks)
        >> resync
        >> done
    )
